declare type Type = 'date' | 'timestamp' | 'number' | 'string' | 'file' | 'boolean' | 'cents' | 'any';

declare class Field {
    type: Type;
    value: any;
    errors: Array<string> | null;
    originalValue: any;
    constructor(value: any, type: Type);
    /**
     * Converts the incoming value
     * before being persisted in the field
     */
    _mapIn(value: any | null): any;
    /**
     * Converts back the value from the field
     */
    _mapOut(): any | null;
    get isDirty(): boolean;
    get hasValue(): boolean;
    mapAndSet(value: any): void;
    set(value: any): void;
    clean(): void;
    reset(): void;
    setErrors(errors: Array<string> | null): void;
}

declare type SaveOptions = {
    patch?: boolean;
};
declare type Schema = {
    [key: string]: any;
};
declare type Values = {
    [key: string]: any;
};
declare type Errors = {
    [key: string]: Array<string> | {};
};
interface Collection {
    create(data: Values): Promise<any>;
}
interface Model {
    save(data: Values, options: SaveOptions): Promise<any>;
}
declare class Form {
    fields: {
        [key: string]: Field;
    };
    constructor(values: Values, schema: Schema);
    get serialized(): Values;
    data(): Values;
    has(attribute: string): boolean;
    get(attribute: string): Field;
    /**
     * Cleans all the forms by changing their original
     * values to correspond to their current values
     */
    cleanAll(): void;
    /**
     * Reset all the forms to their original
     * values thereby discarding all changes made
     */
    resetAll(): void;
    /**
     * Resets all the error fields
     */
    resetErrors(): void;
    /**
     * Sets the errors with a given
     * hash of attribute -> error
     */
    setErrors(errors: Errors): void;
    /**
     * Checks if all fields have a value.
     */
    get isComplete(): boolean;
    /**
     * Return the values from the form as they are
     */
    get values(): Values;
    /**
     * Return whether the form is dirty
     */
    get isDirty(): boolean;
    /**
     * Reset values
     */
    resetValues(): void;
    /**
     * Sets the fields values with a given
     * hash of attribute -> value
     */
    setValues(values: Values): void;
    /**
     * Creates a new model on the given collection
     */
    create(collection: Collection): Promise<any>;
    /**
     * Saves the model with the given fields
     */
    save(model: Model, options?: SaveOptions): Promise<any>;
    handleErrors(fn: () => Promise<any>): Promise<any>;
    /**
     * whether if any of the fields have an error
     */
    get hasErrors(): Boolean;
    /**
     * returns an array of strings with the keys
     * of the fields that are dirty
     */
    get dirtyFieldsKeys(): string[];
    /**
     * returns an array of strings with the keys
     * of the fields that has value
     */
    get fieldsWithValueKeys(): string[];
}

/**
 * Support any input that may contains decimals
 * and other stuff. Check the test cases to understand
 * which cases we handle.
 *
 * Improve this if you find how.
 */
declare function numberParser(str: string): number | null;

export { Field, Form, Type, numberParser };
